"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""


if __name__ == '__main__':
    from . import Loader

    print("Available loaders:")
    for type_name in sorted(Loader.type_names):
        print(f'  - {type_name}:')
        for prio, class_name, cls in Loader.get_all(type_name):
            print(f'    - {class_name:30s} [{prio}]')
